from pathlib import Path
from textwrap import dedent
import sys
from joblib import dump, load
import pkg_resources
from collections import defaultdict

import cv2
from sklearn.neighbors import KNeighborsClassifier
import numpy as np

from facelib import facerec
from facelib._utils import helper

def plot(img, bboxes, landmarks=None, persons_id=None):
    bboxes = bboxes.reshape(-1, 2, 2) * img.shape[:2]
    bboxes = bboxes[...,::-1].astype(np.int32)
    img = img.copy()
    for bbox in bboxes:
        cv2.rectangle(img, tuple(bbox[0]), tuple(bbox[1]), (0,255,0), 2)
    if landmarks is not None:
        landmarks = landmarks.reshape(-1, 5, 2) * img.shape[:2]
        landmarks = landmarks[...,::-1].astype(np.int32)
        for landmark in landmarks:
            for lm in landmark:
                cv2.circle(img, tuple(lm), 1, (0,0,255), 2)
    if persons_id is not None:
        font_face = cv2.FONT_HERSHEY_PLAIN
        font_scale = min(img.shape[:2]) / 500
        thickness = 1
        if bboxes is not None:
            for bbox, p_id in zip(bboxes, persons_id):
                size_text = cv2.getTextSize(p_id, font_face, font_scale, thickness)
                w, h = size_text[0]
                diff_w = int((w - (bbox[1][0] - bbox[0][0])) / 2)
                org = tuple(bbox[0] + [-diff_w, h])
                cv2.putText(img, p_id, org, font_face, font_scale,
                            (255,255,255), thickness, cv2.LINE_AA)
    return img

def get_default_pipeline_detector():
    """Return default pipeline(facelib.facerec.Pipeline)."""
    list_pipeline = helper.get_default_pipeline()
    print('Current pipeline: {}'.format(', '.join(list_pipeline)))
    name_fd, name_ld, name_fe = list_pipeline
    # Face Detector
    if name_fd == 'haarcascade':
        face_detector = facerec.HAARFaceDetector()
    elif name_fd == 'lbpcascade':
        face_detector = facerec.LBPFaceDetector()
    elif name_fd.rsplit('_', 1)[0] == 'ssd_int8':
        tpu = (name_fd.rsplit('_', 1)[-1] == 'tpu')
        face_detector = facerec.SSDFaceDetector(tpu=tpu)
    else:
        assert False, 'Failed to load face detection model.'
    # Landmark Detector
    landmark_detector = facerec.LandmarkDetector(name_ld)
    feature_extractor = facerec.FeatureExtractor(name_fe)
    pipeline = facerec.Pipeline(face_detector, landmark_detector, feature_extractor)
    return pipeline

def gen_folder(posix_path_folder):
    """Yield immediate subdirectories inside a folder."""
    for posix_path in posix_path_folder.iterdir():
        if posix_path.is_dir():
            yield posix_path

def gen_image(posix_path_folder):
    """Yield image files inside a folder."""
    extensions = ['bmp', 'dib', 'jpg', 'jpeg', 'jpe', 'png']
    extensions = ['*.' + ext for ext in extensions]
    for ext in extensions:
        for posix_path in posix_path_folder.glob(ext):
            path_img = str(posix_path)
            name_img = posix_path.name
            img = cv2.imread(path_img)[...,::-1] # bgr->rgb
            yield img, name_img

def augment_data(data, number, noise_range=1e-2):
    """Augments a list/array of lists/arrays."""
    std = np.std(data, axis=0)
    num_augments = number - len(data)
    if num_augments <= 0:
        return data
    diminish = noise_range / float(num_augments)
    augmented = data.copy()
    for _ in range(num_augments):
        rand = np.random.rand(std.shape[0]) * 2 - 1.
        noise = noise_range * rand
        fake_data = noise * std + data[np.random.randint(len(data))]
        fake_data = np.expand_dims(fake_data, axis=0)
        augmented = np.append(augmented, fake_data, axis=0)
        noise_range -= diminish
    return augmented

def train(detector, posix_path_folder, name_clf):
    """Train a classifier using image folders named as people names."""
    dict_features = defaultdict(lambda: np.empty((0, 128), np.float32))
    last_ftr = None
    for posix_path in gen_folder(posix_path_folder):
        name_person = posix_path.stem
        for img, _ in gen_image(posix_path):
            _, _, features = detector.predict(img)
            if len(features) == 0:
                continue
            if len(features) == 1:
                ftr = features[0]
                last_ftr = ftr
            if len(features) > 1:
                if last_ftr is not None:
                    distances = np.linalg.norm(features - last_ftr, axis=1)
                    closest_index = distances.argmin()
                    ftr = features[closest_index]
                else:
                    continue
            dict_features[name_person] = np.append(
                dict_features[name_person], np.expand_dims(ftr, 0), axis=0)
    X = []
    y = []
    for name in dict_features:
        num_augment = helper.get_num_augment()
        features = dict_features[name]
        num_augment = num_augment - len(features)
        for feature in augment_data(features, num_augment):
            X.append(feature)
            y.append(name)
    min_data_per_class = min(np.unique(y, return_counts=True)[1])
    n_neighbors = 3 if min_data_per_class < 5 else 5
    clf = KNeighborsClassifier(metric='cosine', n_neighbors=n_neighbors)
    clf.fit(X, y)
    helper.save_classifier(clf, name_clf)

def predict(detector, posix_path_folder, name_clf):
    config_settings = helper.get_settings_config()
    tolerance = float(config_settings['Predict']['tolerance'])
    clf = helper.get_classifier(name_clf)
    print('Predictions:')
    for img, name_img in gen_image(posix_path_folder):
        _, _, features = detector.predict(img)
        probas = clf.predict_proba(features)
        names_predicted = [
            clf.classes_[p.argmax()] if p.max() > (1 - tolerance) else None for p in probas]
        print('{}: {}'.format(name_img, names_predicted))


"""Image manipulating functions/classes compatible with tensorflow/keras."""

import math
import os

import cv2
import numpy as np
import tensorflow as tf
from skimage import transform as skitrans


class ImageAugmentor:
    """Image augmentation with tensorflow(graph compatible)."""

    def __init__(self,
                 translation_range=(-0., 0.),
                 scale_range=(-0., 0.),
                 rotation_range=(-180, 180)):
        self.tr = tf.constant(translation_range, tf.float32)
        self.sr = tf.constant(scale_range, tf.float32)
        self.rr = tf.constant(rotation_range, tf.float32)
    # performance comparison of augment methods:
    # tests done in 64 batches using tfrecord iterator on i7-6700k
    #   opencv:     183 ms
    #   skimage:    314 ms
    #   keras:      410 ms
    #   scipy:      1410ms

    def augment_opencv(self, img, coordinates):
        """Augments the input image using class parameters.
        
        Notes
        -----
        Modifies coordinates to match with image augmentation.

        """
        img_shape = tf.shape(img)
        img_center = tf.cast(img_shape[:2], tf.float32) / 2.
        translate = tf.random.uniform([2], self.tr[0], self.tr[1]) * tf.cast(img_shape[:2][::-1], tf.float32)
        scale = 1 + tf.random.uniform([2], self.sr[0], self.sr[1])
        pi = tf.constant(math.pi)
        angle = tf.random.uniform([1], self.rr[0], self.rr[1])[0]
        angle = angle * pi / 180
        translate = tf.stack([translate[0], translate[1]])
        center_shift = tf.stack([
            [1., 0, img_center[1]],
            [0, 1, img_center[0]],
            [0, 0, 1]])
        scale_matrix = tf.stack([
            [scale[0], 0, 0.],
            [0, scale[1], 0],
            [0, 0, 1]])
        rotat_matrix = tf.stack([
            [tf.cos(angle), tf.sin(angle), 0.],
            [-tf.sin(angle), tf.cos(angle), 0],
            [0, 0, 1]])
        trans_matrix = tf.stack([
            [1., 0, translate[0]-img_center[1]],
            [0, 1, translate[1]-img_center[0]],
            [0, 0, 1]])
        M = tf.matmul(center_shift, scale_matrix)
        M = tf.matmul(M, rotat_matrix)
        M = tf.matmul(M, trans_matrix)
        M = M[:2]
        coor_shape = tf.shape(coordinates)
        coordinates = tf.reshape(coordinates, [1, -1, 2])
        coordinates = tf.reverse(coordinates, [-1])

        def transform(img, coor, M, img_shape):
            coor = cv2.transform(coor, M)
            img = cv2.warpAffine(img, M, tuple(img_shape), borderMode=cv2.BORDER_REPLICATE)
            return img, coor

        img, coordinates = tf.numpy_function(
            transform,
            [img, coordinates, M, img_shape[:2][::-1]],
            [tf.uint8, tf.float32])
        img = tf.reshape(img, img_shape)
        coordinates = tf.reverse(coordinates, [-1])
        coordinates = tf.reshape(coordinates, coor_shape)
        return (img, coordinates)

def align_custom_numpy(img, landmarks):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    img_shape = np.shape(img)
    landmarks = np.reshape(landmarks, [-1, 2])
    delta_eye = landmarks[1] - landmarks[0]
    angle = np.arctan2(delta_eye[0], delta_eye[1])
    angle = angle * 180 / math.pi
    eye_center = np.mean(landmarks[:2], axis=0)
    mouth_center = np.mean(landmarks[3:], axis=0)
    center = np.mean(
        np.concatenate([landmarks[:2], landmarks[3:]], 0), 0)
    reflect_nosetip = center * 2 - landmarks[2]
    center = (center + reflect_nosetip * [0, 1]) / [1, 2]
    ed_eye = np.linalg.norm(delta_eye) # euclidean distance between eyes
    ed_eye_mouth = np.linalg.norm(eye_center - mouth_center)
    yx = center - np.stack([ed_eye_mouth, ed_eye])
    tmp = np.abs(center - landmarks[2])
    width = np.max([ed_eye, tmp[0]], 0) * 2
    height = np.max([ed_eye_mouth, tmp[1]], 0) * 2
    bbox = np.stack([yx[0], yx[1], height, width])
    bbox = bbox.astype(np.int32)

    M = cv2.getRotationMatrix2D(tuple(center[::-1]), angle, 1)
    img = cv2.warpAffine(img, M, (img_shape[1], img_shape[0]))
    y, x, h, w = (bbox[0], bbox[1], bbox[2], bbox[3])
    h = h + y
    w = w + x
    y = np.max([y, 0])
    y = np.min([y, img_shape[0] - 1])
    x = np.max([x, 0])
    x = np.min([x, img_shape[1] - 1])
    h = np.max([h - y, 1])
    h = np.min([h, img_shape[0] - y])
    w = np.max([w - x, 1])
    w = np.min([w, img_shape[1] - x])
    dst = img[y:y+h, x:x+w, :]
    return dst

def align_custom_tensorflow(img, landmarks):
    """Tensorflow graph compatible version of method: align_custom_numpy."""
    img_shape = tf.shape(img)
    landmarks = tf.reshape(landmarks, [-1, 2])
    pi = tf.constant(math.pi)
    delta_eye = landmarks[1] - landmarks[0]
    angle = tf.atan2(delta_eye[0], delta_eye[1])# in radians
    angle = angle * 180 / pi
    eye_center = tf.reduce_mean(landmarks[:2], axis=0)
    mouth_center = tf.reduce_mean(landmarks[3:], axis=0)
    center = tf.reduce_mean(
        tf.concat([landmarks[:2], landmarks[3:]], 0), 0)
    reflect_nosetip = center * 2 - landmarks[2]
    center = (center + reflect_nosetip * [0, 1]) / [1, 2]
    ed_eye = tf.linalg.norm(delta_eye) #euclidean distance
    ed_eye_mouth = tf.linalg.norm(eye_center - mouth_center)
    yx = center - tf.stack([ed_eye_mouth, ed_eye])
    tmp = tf.math.abs(center - landmarks[2])
    width = tf.reduce_max([ed_eye, tmp[0]], 0) * 2
    height = tf.reduce_max([ed_eye_mouth, tmp[1]], 0) * 2
    bbox = tf.stack([yx[0], yx[1], height, width])
    bbox = tf.cast(bbox, tf.int32)
    def rotate(img, center, angle, shape):
        M = cv2.getRotationMatrix2D(tuple(center), angle, 1)
        dst = cv2.warpAffine(img, M, (shape[0], shape[1]))
        return dst
    img = tf.numpy_function(
        rotate,
        [img, center[::-1], angle, img_shape[:2][::-1]],
        tf.uint8)
    img = tf.reshape(img, img_shape)
    y, x, h, w = (bbox[0], bbox[1], bbox[2], bbox[3])
    h = h + y
    w = w + x
    y = tf.reduce_max([y, 0])
    y = tf.reduce_min([y, img_shape[0] - 1])
    x = tf.reduce_max([x, 0])
    x = tf.reduce_min([x, img_shape[1] - 1])
    h = tf.reduce_max([h - y, 1])
    h = tf.reduce_min([h, img_shape[0] - y])
    w = tf.reduce_max([w - x, 1])
    w = tf.reduce_min([w, img_shape[1] - x])
    bbox = tf.stack([y, x, h, w], axis=0)
    dst = tf.image.crop_to_bounding_box(
        img, bbox[0], bbox[1], bbox[2], bbox[3])
    return dst

def align_golden_ratio_numpy(img, landmarks):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    img_shape = np.shape(img)
    landmarks = np.reshape(landmarks, [-1, 2])
    delta_eye = landmarks[1] - landmarks[0]
    rad = np.arctan2(delta_eye[0], delta_eye[1])
    angle = rad * 180 / math.pi
    ed_eye = np.linalg.norm(delta_eye) # euclidean distance between eyes
    eye_center = np.mean(landmarks[:2], axis=0)
    center = np.array(img_shape[:2]) / 2
    center_shift = eye_center - center
    width = ed_eye * 1.865
    height = ed_eye * 3.018
    rotation_matrix = cv2.getRotationMatrix2D(tuple(eye_center[::-1]), angle, 1)
    rotation_matrix = np.concatenate([rotation_matrix, [[0,0,1]]], 0)
    translation_matrix = np.array([
        [1, 0, center_shift[1]],
        [0, 1, center_shift[0] + ed_eye * 0.3],
        [0, 0, 1]])
    translation_matrix = np.linalg.inv(translation_matrix)
    M = np.matmul(rotation_matrix, translation_matrix)
    M = M[:2]
    img = cv2.warpAffine(img, M, (img_shape[1], img_shape[0]))
    bbox = np.array([center[0]-height/2, center[1]-width/2, height, width], dtype=np.int32)
    y, x, h, w = bbox
    h = h + y
    w = w + x
    y = np.max([y, 0])
    y = np.min([y, img_shape[0] - 1])
    x = np.max([x, 0])
    x = np.min([x, img_shape[1] - 1])
    h = np.max([h - y, 1])
    h = np.min([h, img_shape[0] - y])
    w = np.max([w - x, 1])
    w = np.min([w, img_shape[1] - x])
    dst = img[y:y+h, x:x+w, :]
    return dst

def align_golden_ratio_tensorflow(img, landmarks):
    """Tensorflow graph compatible version of method: align_golden_ratio_numpy."""
    img_shape = tf.shape(img)
    landmarks = tf.reshape(landmarks, [-1, 2])
    pi = tf.constant(math.pi)
    delta_eye = landmarks[1] - landmarks[0]
    rad = tf.atan2(delta_eye[0], delta_eye[1])
    angle = rad * 180 / pi
    ed_eye = tf.linalg.norm(delta_eye) # euclidean distance between eyes
    eye_center = tf.reduce_mean(landmarks[:2], axis=0)
    center = tf.cast(img_shape[:2], tf.float32) / 2
    center_shift = eye_center - center
    width = ed_eye * 1.865
    height = ed_eye * 3.018
    translation_matrix = tf.convert_to_tensor([
        [1, 0, center_shift[1]],
        [0, 1, center_shift[0] + ed_eye * 0.3],
        [0, 0, 1]])
    translation_matrix = tf.linalg.inv(translation_matrix)
    def fn(img, center, eye_center, angle, img_shape, translation_matrix):
        rotation_matrix = cv2.getRotationMatrix2D(tuple(eye_center[::-1]), angle, 1)
        rotation_matrix = np.concatenate([rotation_matrix, [[0,0,1]]], 0)
        M = np.matmul(rotation_matrix, translation_matrix)
        M = M[:2]
        img = cv2.warpAffine(img, M, (img_shape[1], img_shape[0]))
        return img
    img = tf.numpy_function(
        fn,
        [img, center, eye_center, angle, img_shape, translation_matrix],
        tf.uint8)
    img = tf.reshape(img, img_shape)
    bbox = tf.convert_to_tensor(
        [center[0]-height/2, center[1]-width/2, height, width], dtype=tf.int32)
    y, x, h, w = bbox[0], bbox[1], bbox[2], bbox[3]
    h = h + y
    w = w + x
    y = tf.reduce_max([y, 0])
    y = tf.reduce_min([y, img_shape[0] - 1])
    x = tf.reduce_max([x, 0])
    x = tf.reduce_min([x, img_shape[1] - 1])
    h = tf.reduce_max([h - y, 1])
    h = tf.reduce_min([h, img_shape[0] - y])
    w = tf.reduce_max([w - x, 1])
    w = tf.reduce_min([w, img_shape[1] - x])
    dst = img[y:y+h, x:x+w, :]
    return dst

def align_similarity_numpy(img, lm):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    lm = np.reshape(lm, [-1, 2])
    img_shape = img.shape[:2]
    dst = np.array([
        [0.315, 0.46],
        [0.685, 0.46],
        [0.50,  0.64],
        [0.35,  0.825],
        [0.65,  0.825]], dtype=np.float32 )
    dst  = dst * img_shape[::-1]
    lm = np.flip(lm, -1)
    M = cv2.estimateAffinePartial2D(lm, dst)[0]
    warped = cv2.warpAffine(img,M,(img_shape[1],img_shape[0]), borderValue = 0.0)
    return warped

def align_similarity_tensorflow(img, lm):
    """Tensorflow graph compatible version of method: align_similarity_numpy."""
    img_shape = tf.shape(img)
    dst = tf.numpy_function(align_similarity_numpy, [img, lm], tf.uint8)
    dst = tf.reshape(dst, img_shape)
    return dst

def align_similarity_zoomed_numpy(img, lm):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    lm = np.reshape(lm, [-1, 2])
    img_shape = img.shape[:2]
    dst = np.array([
        [0.222, 0.37],
        [0.777, 0.37],
        [0.50,  0.64],
        [0.275, 0.9175],
        [0.275, 0.9175]], dtype=np.float32 )
    dst  = dst * img_shape[::-1]
    lm = np.flip(lm, -1)
    M = cv2.estimateAffinePartial2D(lm, dst)[0]
    warped = cv2.warpAffine(img,M,(img_shape[1],img_shape[0]), borderValue = 0.0)
    return warped

def align_similarity_zoomed_tensorflow(img, lm):
    """Tensorflow graph compatible version of method: align_similarity_zoomed_numpy."""
    img_shape = tf.shape(img)
    dst = tf.numpy_function(align_similarity_zoomed_numpy, [img, lm], tf.uint8)
    dst = tf.reshape(dst, img_shape)
    return dst

def align_similarity_v2_numpy(img, lm):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    lm = np.reshape(lm, [-1, 2])
    lm = lm[...,::-1]
    img_shape = img.shape[:2]
    dst = np.array([
        [0.34, 0.46],
        [0.66, 0.46],
        [0.5, 0.64],
        [0.37, 0.82],
        [0.63, 0.82] ], dtype=np.float32 )
    dst =  dst * img_shape[::-1]
    tform = skitrans.SimilarityTransform()
    tform.estimate(lm, dst)
    M = tform.params[0:2,:]
    warped = cv2.warpAffine(img,M,(img_shape[1],img_shape[0]), borderValue = 0.)
    return warped

def align_insightface_tensorflow(img, lm):
    """Tensorflow graph compatible version of method: align_insightface_numpy."""
    img_shape = tf.shape(img)
    dst = tf.numpy_function(align_insightface_numpy, [img, lm], tf.uint8)
    dst = tf.reshape(dst, img_shape)
    return dst

def align_insightface_numpy(img, lm):
    """Image aligning function implemented in numpy.

    Notes
    -----
    Landmarks should be:
    [left_eye_y, left_eye_x, right_eye_y, right_eye_x,
    nosetip_y, nosetip_x, mouth_left_edge_y, mouth_left_edge_x,
    mouth_right_edge_y, mouth_right_edge_x]
    """
    lm = np.reshape(lm, [-1, 2])
    lm = lm[...,::-1]
    img_shape = img.shape[:2]
    dst = np.array([
        [0.34191608, 0.46157411],
        [0.65653392, 0.45983393],
        [0.500225  , 0.64050538],
        [0.3709759 , 0.82469198],
        [0.631517  , 0.82325091]])
    dst =  dst * img_shape[::-1]
    tform = skitrans.SimilarityTransform()
    tform.estimate(lm, dst)
    M = tform.params[0:2,:]
    warped = cv2.warpAffine(img,M,(img_shape[1],img_shape[0]), borderValue = 0.)
    return warped

def align_similarity_v2_tensorflow(img, lm):
    """Tensorflow graph compatible version of method: align_similarity_v2_numpy."""
    img_shape = tf.shape(img)
    dst = tf.numpy_function(align_similarity_v2_numpy, [img, lm], tf.uint8)
    dst = tf.reshape(dst, img_shape)
    return dst



"""Helper functions for facelib module."""
import platform
import subprocess
import sys
from configparser import ConfigParser, ExtendedInterpolation
from pathlib import Path
from urllib import request
from textwrap import dedent

import pkg_resources
from joblib import dump, load


# Tensorflow lite runtime
def get_tflite_runtime_link():
    """Get tflite_runtime whl package from google."""
    # Get system info
    assert_msg = 'tflite_runtime is not available for this platform'
    p_system = platform.system()
    if p_system == 'Linux':
        plt = 'linux'
    elif p_system == 'Windows':
        plt = 'win'
    elif p_system == 'Darwin':
        assert platform.mac_ver()[0][:5] == '10.14', assert_msg
        plt = 'macosx_10_14'
    else:
        sys.exit(assert_msg)

    # Get python version
    major, minor = platform.python_version_tuple()[:2]
    python_version = major + minor

    # Get processor architecture
    arch = platform.machine().lower()

    # Assertion
    assert python_version in ['35', '36', '37'] and arch in ['armv7l', 'aarch64', 'x86_64', 'amd64']

    url_base = ('https://dl.google.com/coral/python/'
                'tflite_runtime-2.1.0.post1-cp{0}-cp{0}m-{1}_{2}.whl')
    url_whl = url_base.format(python_version, plt, arch)
    return url_whl

def install_tflite_runtime():
    """Download tflite-runtime pip package."""
    url_whl = get_tflite_runtime_link()
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', url_whl])

# Template
def get_templates_config():
    """Get templates config."""
    name_config = 'templates.ini'
    path_config = pkg_resources.resource_filename('facelib._utils', name_config)
    config = ConfigParser(interpolation=ExtendedInterpolation())
    config.read(path_config)
    return config

def set_templates_config(config):
    """Write config to templates config file."""
    name_config = 'templates.ini'
    path_config = pkg_resources.resource_filename('facelib._utils', name_config)
    with open(path_config, 'w') as configfile:
        config.write(configfile)

# Settings
def get_settings_config():
    """Set facelib module settings."""
    name_config = 'settings.ini'
    path_config = pkg_resources.resource_filename('facelib._utils', name_config)
    config = ConfigParser(interpolation=ExtendedInterpolation())
    config.read(path_config)
    assert 1 >= float(config['Predict']['tolerance']) > 0, "Tolerance should be: (0, 1]"
    return config

def set_settings_config(config):
    """Get config with facelib module settings."""
    name_config = 'settings.ini'
    path_config = pkg_resources.resource_filename('facelib._utils', name_config)
    with open(path_config, 'w') as configfile:
        config.write(configfile)

# Model
def get_links_config():
    """Get config with url links to models datas."""
    name_config = 'links_to_models.ini'
    path_config = pkg_resources.resource_filename('facelib._utils', name_config)
    config = ConfigParser(interpolation=ExtendedInterpolation())
    config.read(path_config)
    assert config, 'Error opening config file: ' + name_config
    return config

def set_default_model(type_model, name_model):
    """Set default model."""
    assert_model_availability(type_model, name_model)
    config = get_templates_config()
    config['CurrentTemplate'][type_model] = name_model
    set_templates_config(config)

def get_default_model(type_model):
    """Get name of the current model."""
    types = ['face_detection', 'landmark_detection', 'feature_extraction']
    assert type_model in types, 'Model type is wrong.'
    config = get_templates_config()
    model = config['CurrentTemplate'][type_model]
    return model

def get_available_models(type_model):
    """Get available model names as list."""
    links = get_links_config()
    available_models = [key for key in links[type_model].keys() if not key.startswith('_')]
    return available_models

def assert_model_availability(type_model, name_model):
    """Assert model availability."""
    types = [
        'face_detection',
        'landmark_detection',
        'feature_extraction',]
    assert type_model in types, 'Available types are: {}'.format(str(types))
    links = get_links_config()
    keys = [key for key in links[type_model].keys() if not key.startswith('_')]
    assert name_model in keys, 'Available {} {} models are:\n{}'.format(
        *type_model.split('_'), str(keys))

def get_path(type_model, name_model, force_download=True):
    """Get path to model data."""
    assert_model_availability(type_model, name_model)
    links = get_links_config()
    link_model_data = links[type_model][name_model]
    filename = link_model_data.rsplit('/', 1)[-1]
    path_file = pkg_resources.resource_filename(
        'facelib.facerec.' + type_model,
        'data/' + filename
    )
    exist = Path(path_file).exists()
    if force_download and not exist:
        install_data(type_model, name_model)
    return path_file

def get_link(type_model, name_model):
    """Get links to model data."""
    links = get_links_config()
    url_link = links[type_model][name_model]
    return url_link

def install_data(type_model, name_model):
    """Install model data."""
    path_file = get_path(type_model, name_model, force_download=False)
    link_data = get_link(type_model, name_model)
    response = request.urlopen(link_data)
    Path(path_file).parent.mkdir(parents=True, exist_ok=True)
    with open(path_file, 'wb') as _f:
        print('Downloading from url <{}>...'.format(link_data))
        _f.write(response.read())
        print('File <{}> successfully created.'.format(path_file))

# Pipeline
def print_templates_info():
    """Print current/available templates."""
    config = get_templates_config()
    current_pipeline = get_default_pipeline()
    message = dedent("""\
    Current pipeline is:
        (F)ace Detection:     {}
        (L)andmark Detection: {}
        (Fe)ature Extraction: {}\n
    Available templates are:""".format(*current_pipeline))
    print(message)
    tmp = '        (F){} (L){} (Fe){}'
    for name_template in config['PipelineTemplates'].keys():
        pipeline = get_pipeline(name_template)
        print('    {}'.format(name_template))
        print(tmp.format(*pipeline))


def get_default_pipeline():
    """Get the default pipeline as list of model names."""
    config = get_templates_config()
    default_config = config['CurrentTemplate']
    _fd = default_config['face_detection']
    _ld = default_config['landmark_detection']
    _fe = default_config['feature_extraction']
    return [_fd, _ld, _fe]

def set_default_pipeline(name_pipeline):
    """Set the current pipeline."""
    _fd, _ld, _fe = get_pipeline(name_pipeline)
    config = get_templates_config()
    default_config = config['CurrentTemplate']
    default_config['face_detection'] = _fd
    default_config['landmark_detection'] = _ld
    default_config['feature_extraction'] = _fe
    set_templates_config(config)

def get_pipeline(name_pipeline):
    """Get list of model names as pipeline."""
    config = get_templates_config()
    str_value = config['PipelineTemplates'][name_pipeline]
    list_value = str_value.split(',')
    return list_value

def del_pipeline(name_pipeline):
    """Delete a pipeline."""
    config = get_templates_config()
    assert name_pipeline in config['PipelineTemplates'].keys(), 'Pipeline does not exist.'
    del config['PipelineTemplates'][name_pipeline]
    set_templates_config(config)

def save_pipeline(list_pipeline, name_pipeline):
    """Name a pipeline and save it."""
    assert len(list_pipeline) == 3, 'Number of models should be 3.'
    # assert if available
    types = ['face_detection', 'landmark_detection', 'feature_extraction']
    for type_model, name_model in zip(types, list_pipeline):
        assert_model_availability(type_model, name_model)
    config = get_templates_config()
    config['PipelineTemplates'][name_pipeline] = ','.join(list_pipeline)
    set_templates_config(config)

# Classification
def save_classifier(clf, name_clf):
    available_names = [x.split('_-_')[0] for x in get_available_classifiers()]
    if name_clf in available_names:
        print('Classifier named `{}` is overwritten.'.format(name_clf))
    name_fe = get_default_model('feature_extraction')
    name_fe = name_fe.split('_', 1)[0]
    path_save = get_classifier_path(name_clf, name_fe)
    dump(clf, str(path_save))
    print('Classifier named `{}` succesfully trained and saved.'.format(name_clf))

def get_classifier(name_clf):
    available_classifiers = get_available_classifiers()
    available_names = [x.split('_-_')[0] for x in available_classifiers]
    if name_clf not in available_names:
        print('Classifier with name `{}` does not exist.'.format(name_clf))
        print('Available classifiers are:')
        print_classifiers_info()
        sys.exit()
    name_fe_default = get_default_model('feature_extraction').split('_', 1)[0]
    index = available_names.index(name_clf)
    name_fe = available_classifiers[index].split('_-_')[1]
    if name_fe != name_fe_default:
        message = ('Error: Current feature extractor `{}` is not compatible with'
                   ' the feature extractor `{}` used to train classifier `{}`.')
        sys.exit(message.format(name_fe_default, name_fe, name_clf))
    posix_path_clf = get_classifier_path(name_clf, name_fe)
    clf = load(str(posix_path_clf))
    return clf

def get_classifier_path(name_clf, name_fe):
    """Get posix path to a classifier."""
    name_save = name_clf + '_-_' + name_fe
    path_classifier = pkg_resources.resource_filename(
        'facelib.facerec', '_dataset/classifier/' + name_save)
    posix_path = Path(path_classifier).with_suffix('.joblib')
    posix_path.parent.mkdir(parents=True, exist_ok=True)
    return posix_path

def get_available_classifiers():
    """Get list of posix paths to classifiers."""
    path_dir_classifier = pkg_resources.resource_filename(
        'facelib.facerec', '_dataset/classifier')
    posix_path = Path(path_dir_classifier)
    return [p.stem for p in posix_path.glob('*.joblib')]

def print_classifiers_info():
    for name in get_available_classifiers():
        name_clf, name_fe = name.split('_-_')
        posix_path_clf = get_classifier_path(name_clf, name_fe)
        clf = load(str(posix_path_clf))
        classes = clf.classes_
        print(dedent("""
            Classifier name: {}
                Feature extractor used is {}
                Classes: {}""").format(name_clf, name_fe, ', '.join(classes)))

# Training
def get_num_augment():
    """Get min number of samples per image to train."""
    config = get_settings_config()
    num_augment = int(config['Train']['num_augment'])
    return num_augment



"""Keras implementation of various loss functions."""

import tensorflow as tf


def rmse(y_true, y_pred):
    """Keras implementation of root mean squared error."""
    loss = tf.losses.mse(y_true, y_pred)
    loss = tf.sqrt(loss)
    loss = tf.reduce_mean(loss)
    return loss



"""Cascaded face detection using OpenCv."""

from pathlib import Path

import cv2
import numpy as np
import pkg_resources

from facelib._utils import helper

class CascadeClassifier:
    """OpenCV cascade classifier.

    
    """
    def __init__(self, path_xml, min_neighbors, scale, resize=None, input_img_channel='rgb'):
        self.face_cascade = cv2.CascadeClassifier(path_xml)
        self.min_neighbors = min_neighbors
        self.resize = resize
        assert input_img_channel in ['gray', 'bgr', 'rgb']
        self.input_img_channel = input_img_channel
        self.scale = scale

    def predict(self, img):
        if self.input_img_channel != 'gray':
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if self.resize is not None:
            img = cv2.resize(img, self.resize)
        bboxes = self.face_cascade.detectMultiScale(img, scaleFactor=self.scale, minNeighbors=self.min_neighbors)
        bboxes = np.array(bboxes)
        bboxes = bboxes.reshape(-1, 2, 2)[:,:,::-1]
        top_left = bboxes[:,0]
        bottom_right = np.sum(bboxes, -2)
        bboxes = np.concatenate([top_left, bottom_right], -1)
        bboxes = bboxes.reshape(-1, 2, 2)
        bboxes = bboxes / img.shape[:2]
        return bboxes


class LBPCascade(CascadeClassifier):
    def __init__(self, min_neighbors=3, scale=1.2, **kwargs):
        name_model = 'lbpcascade'
        type_model = 'face_detection'
        path_xml = helper.get_path(type_model, name_model)
        super(LBPCascade, self).__init__(
            path_xml,
            min_neighbors=min_neighbors,
            scale=scale,
            **kwargs,
        )

class HAARCascade(CascadeClassifier):
    def __init__(self, min_neighbors=5, scale=1.2, **kwargs):
        name_model = 'haarcascade'
        type_model = 'face_detection'
        path_xml = helper.get_path(type_model, name_model)
        super(HAARCascade, self).__init__(
            path_xml,
            min_neighbors=min_neighbors,
            scale=scale,
            **kwargs,
        )



"""Entry point for facelib pip package."""
import argparse
import sys
from itertools import groupby
from pathlib import Path
from textwrap import dedent

import cv2


from facelib._utils import helper

try:
    from facelib._utils import face_recognition
except ImportError:
    print('Warning: Failed to load FaceRecognition.\n'
          'try installing tflite-runtime:\n'
          'foo@bar:~$ python3 -m facelib install --tflite-runtime')

class FaceLib:
    def __init__(self,):
        self.parser = argparse.ArgumentParser(
            description='Face Recognition Library',
            formatter_class=argparse.RawTextHelpFormatter)
        self.parser.set_defaults(func=lambda args: self.parser.print_help())
        self.subparsers = self.parser.add_subparsers()
        self._parse()


    def _parse(self,):
        args = sys.argv[1:]
        self._parser_predict()
        self._parser_train()
        self._parser_model()
        self._parser_settings()
        self._parser_install()
        self._parser_info()
        parsed = self.parser.parse_args(args)
        parsed.func(parsed)

    def _parser_settings(self,):
        config = helper.get_settings_config()
        parser = self.subparsers.add_parser(name='settings',
                                            help='adjust settings',
                                            formatter_class=argparse.RawTextHelpFormatter)
        # Group: Predict
        config_predict = config['Predict']
        group_predict = parser.add_argument_group('Predict')
        group_predict.add_argument('-t', '--tolerance', type=float, help=dedent("""\
            Classification tolerance. range: (0:1]
            current:{}, recommended:0.5""".format(config_predict['tolerance'])))
        # Group: Train
        config_train = config['Train']
        group_train = parser.add_argument_group('Train')
        group_train.add_argument('-n', '--num-augment', type=int, help=dedent("""\
            Augment class images.
            current:{}, recommended:5""".format(config_train['num_augment'])))
        parser.set_defaults(func=self._settings)

    def _parser_train(self,):
        parser = self.subparsers.add_parser(name='train', help=(dedent("""\
                train a face classifier using images inside a folder
                training_folder/
                ├── alice/
                │   ├── img1.jpg
                ├── bob/
                │   ├── img1.jpg
                │   ├── img2.jpg""")))
        parser.add_argument('path', type=str, help='folder to be predicted')
        parser.add_argument('classifier', type=str, help='name of the classifier to be trained')
        parser.set_defaults(func=self._train)

    def _parser_predict(self,):
        parser = self.subparsers.add_parser(name='predict',
                                            help='make predictions on images inside a folder')
        parser.add_argument('path', type=str, help='folder to be predicted')
        parser.add_argument('-clf', '--classifier', type=str,
                            help=('predict person ids'))
        parser.add_argument('-p', '--plot', action='store_true',
                            help='plot and save images to folder: (path)_facelib_plotted/')
        parser.add_argument('-c', '--crop', action='store_true',
                            help='crop faces and save to folder: (path)_facelib_cropped/')
        parser.set_defaults(func=self._predict)

    def _parser_model(self,):
        def _normalize(list_model):
            grouped_list = [list(g) for k, g in groupby(list_model, key=lambda x: x.split('_')[0])]
            normalized = '{'
            for group in grouped_list:
                normalized += ', '.join(group)
                normalized += ',\n'
            normalized = normalized[:-2] + '}'
            return normalized
        parser = self.subparsers.add_parser(name='model', help='select/delete/load/save models')
        parser.set_defaults(func=lambda args: parser.print_help())
        subparsers_model = parser.add_subparsers()
        # Parser: select
        parser_select = subparsers_model.add_parser(name='select', help='select default models',
                                                    formatter_class=argparse.RawTextHelpFormatter)
        fd_models = helper.get_available_models('face_detection')
        parser_select.add_argument('-fd', '--face-detection', type=str, choices=fd_models,
                                   metavar='', help=_normalize(fd_models))
        lm_models = helper.get_available_models('landmark_detection')
        parser_select.add_argument('-ld', '--landmark-detection', type=str, choices=lm_models,
                                   metavar='', help=_normalize(lm_models))
        fe_models = helper.get_available_models('feature_extraction')
        parser_select.add_argument('-fe', '--feature-extraction', type=str, choices=fe_models,
                                   metavar='', help=_normalize(fe_models))
        parser_select.set_defaults(func=self._model_select)
        # Parser: load
        parser_load = subparsers_model.add_parser(name='load', help='select a pipeline template')
        parser_load.add_argument('name', type=str, help='name of the model')
        parser_load.set_defaults(func=self._model_load)
        # Parser: save
        parser_save = subparsers_model.add_parser(name='save', help='save current pipeline')
        parser_save.add_argument('name', type=str, help='name of the model')
        parser_save.set_defaults(func=self._model_save)
        # Parser: delete
        parser_delete = subparsers_model.add_parser(name='delete',
                                                    help='delete a pipeline template')
        parser_delete.add_argument('name', type=str, help='name of the model')
        parser_delete.set_defaults(func=self._model_delete)

    def _parser_install(self,):
        parser = self.subparsers.add_parser(name='install',
                                            help='install a package or dataset')
        parser.add_argument('-tr', '--tflite-runtime', action='store_true',
                            help='install tflite-runtime(pip package)')
        parser.set_defaults(func=self._install)

    def _parser_info(self,):
        parser = self.subparsers.add_parser(name='info',
                                            help='get information on models/classifiers')
        # Group: Classification
        group_classification = parser.add_argument_group('Classification')
        group_classification.add_argument('-clf', '--classifier', action='store_true')
        # Group: Pipeline
        group_pipeline = parser.add_argument_group('Pipeline')
        group_pipeline.add_argument('-t', '--template', action='store_true')
        parser.set_defaults(func=self._info)

    @staticmethod
    def _settings(args):
        if (args.tolerance and args.num_augment) is None:
            print('Warning: No arguments recieved.')
        config = helper.get_settings_config()
        # Group: Predict
        if args.tolerance is not None:
            config['Predict']['tolerance'] = str(args.tolerance)
        # Group: Train
        if args.num_augment is not None:
            config['Train']['num_augment'] = str(args.num_augment)
        helper.set_settings_config(config)

    @staticmethod
    def _train(args):
        posix_path_folder = Path(args.path)
        name_clf = args.classifier
        detector = face_recognition.get_default_pipeline_detector()
        face_recognition.train(detector, posix_path_folder, name_clf)

    @staticmethod
    def _predict(args):
        detector = face_recognition.get_default_pipeline_detector()
        posix_path_folder = Path(args.path) # test folder
        config_settings = helper.get_settings_config()
        # Load classifier(person id)
        if args.classifier is not None:
            tolerance = float(config_settings['Predict']['tolerance'])
            clf = helper.get_classifier(args.classifier)
        # Walk through the images inside the test folder
        for img, name_img in face_recognition.gen_image(posix_path_folder):
            bboxes, landmarks, features = detector.predict(img)
            print(name_img)
            print('├───{} faces detected'.format(len(bboxes)))
            # Predict person ids and verbose to terminal
            if args.classifier is not None:
                probas = clf.predict_proba(features)
                names_predicted = [
                    clf.classes_[p.argmax()] if p.max() > (1 - tolerance) else None for p in probas]
                print('├───{}'.format(names_predicted))
            # Plot and save test images
            if args.plot:
                path_plot = posix_path_folder.parent/(
                    posix_path_folder.stem + '_facelib_plotted')
                path_plot.mkdir(parents=True, exist_ok=True)
                path_plotted_img = path_plot/name_img
                if args.classifier is None:
                    names_predicted = None
                img_plotted = face_recognition.plot(img, bboxes, landmarks, names_predicted)
                cv2.imwrite(str(path_plotted_img), img_plotted[...,::-1])
            # Crop and save test faces
            if args.crop:
                path_crop = posix_path_folder.parent/(
                    posix_path_folder.stem + '_facelib_cropped')
                path_crop.mkdir(parents=True, exist_ok=True)
                faces_cropped = [detector.crop_to_bbox(img, bbox) for bbox in bboxes]
                if args.classifier is not None:
                    i = 0
                    for img, name in zip(faces_cropped, names_predicted):
                        if name is None:
                            name = '__Unknown__'
                        path_img = path_crop/name/(str(i) + '_' + name_img)
                        path_img.parent.mkdir(parents=True, exist_ok=True)
                        cv2.imwrite(str(path_img), img[...,::-1])
                        i += 1
                else:
                    i = 0
                    for img in faces_cropped:
                        path_img = path_crop/(str(i) + '_' + name_img)
                        cv2.imwrite(str(path_img), img[...,::-1])
                        i += 1

    @staticmethod
    def _model_select(args):
        current_pipeline = helper.get_default_pipeline()
        if args.face_detection is not None:
            helper.set_default_model('face_detection', args.face_detection)
        if args.landmark_detection is not None:
            helper.set_default_model('landmark_detection', args.landmark_detection)
        if args.feature_extraction is not None:
            helper.set_default_model('feature_extraction', args.feature_extraction)
        new_pipeline = helper.get_default_pipeline()
        if current_pipeline == new_pipeline:
            print('WARNING: No updates made...')
        print('Face Recognition: {}\nLandmark Detection: {}\nFeature Extraction: {}'.format(
            *new_pipeline))

    @staticmethod
    def _model_load(args):
        helper.set_default_pipeline(args.name)

    @staticmethod
    def _model_save(args):
        pipeline = helper.get_default_pipeline()
        helper.save_pipeline(pipeline, args.name)
        print('Pipeline: [{}] saved as {}'.format(', '.join(pipeline), args.name))

    @staticmethod
    def _model_delete(args):
        helper.del_pipeline(args.name)

    @staticmethod
    def _install(args):
        if not args.tflite_runtime:
            print('Warning: No arguments recieved.')
        if args.tflite_runtime:
            helper.install_tflite_runtime()

    @staticmethod
    def _info(args):
        if not (args.classifier and args.template):
            print('Warning: No arguments recieved.')
        if args.classifier:
            helper.print_classifiers_info()
        if args.template:
            helper.print_templates_info()

if __name__ == '__main__':
    FaceLib()



"""Quantized Single Shot Detector for face detection."""
from pathlib import Path

import cv2
import numpy as np
import pkg_resources
from tflite_runtime.interpreter import Interpreter, load_delegate

from facelib._utils import helper


class SSDFaceDetector:
    """SSD face detection(supports tpu).

    Notes
    -----
    ref(tflite) : http://download.tensorflow.org/models/object_detection/facessd_mobilenet_v2_quantized_320x320_open_image_v4.tar.gz

    Warnings
    --------
    If the image to be predicted is not square shaped, prediction accuracy is dramatically reduced.
    """

    def __init__(self, input_img_channel='rgb', resize=None, tpu=False):
        if tpu:
            name_model = 'ssd_int8_tpu'
            delegates = [load_delegate('libedgetpu.so.1')]
        else:
            name_model = 'ssd_int8_cpu'
            delegates = []
        path_tflite = helper.get_path('face_detection', name_model)
        self.face_detection_inference = Interpreter(
            model_path=str(path_tflite),
            experimental_delegates=delegates)
        self.face_detection_inference.allocate_tensors()
        self.index_input = self.face_detection_inference.get_input_details()[0]['index']
        self.index_boxes = self.face_detection_inference.get_output_details()[0]['index']
        self.index_probs = self.face_detection_inference.get_output_details()[2]['index']
        assert input_img_channel in ['bgr', 'rgb'], 'Incorrect input_img_channel'
        self.input_img_channel = input_img_channel

    def _preprocess(self, img):
        if self.input_img_channel == 'bgr':
            img = img[...,::-1] # bgr to rgb conversion
        img = cv2.resize(img, (320, 320))
        img = np.expand_dims(img, 0)
        return img

    @staticmethod
    def _extract_boxes(boxes, probs):
        threshold = 1e-1
        mask = np.less(probs, 1.)
        probs = probs[mask]
        mask = np.greater(probs, threshold)
        num_boxes = np.count_nonzero(mask)
        output = boxes[0][:num_boxes]
        return output

    def predict(self, img):
        """Detect faces inside an image.

        Parameters
        ----------
        img : ndarray
            3d array with shape (height,width,3) with dtype `uint8`.
        """
        original_shape = img.shape
        img = self.squarize_img(img)
        squared_shape = img.shape
        img = self._preprocess(img)
        self.face_detection_inference.set_tensor(self.index_input, img)
        self.face_detection_inference.invoke()
        pred_boxes = self.face_detection_inference.get_tensor(self.index_boxes)
        pred_probs = self.face_detection_inference.get_tensor(self.index_probs)
        bboxes = self._extract_boxes(pred_boxes, pred_probs)
        bboxes = bboxes.reshape(-1, 2, 2)
        fix = bboxes[:,1,1] - bboxes[:,0,1]
        fix = fix / 6.
        fix = np.expand_dims(fix, -1)
        fix = fix * [0., 1, 0, -1]
        fix = fix.reshape(-1, 2, 2)
        bboxes = bboxes + fix
        bboxes = self.normalize_bboxes(bboxes, original_shape, squared_shape)
        return bboxes

    @staticmethod
    def squarize_img(img):
        height, width = img.shape[:2]
        max_dim = max(height, width)
        new_shape = (max_dim, max_dim, 3)
        squarized = np.zeros(new_shape, dtype=img.dtype)
        squarized[:height,:width] += img
        return squarized

    @staticmethod
    def normalize_bboxes(bboxes, original_shape, squared_shape):
        change_ratio = np.array(squared_shape[:2], dtype=bboxes.dtype) / original_shape[:2]
        bboxes = bboxes * change_ratio
        return bboxes



"""Keras implementation of euclidean distance loss functions with various norms."""

import tensorflow as tf


class EuclideanDistance:
    """Euclidean Distance with norms for facial landmark detection.
    
    Parameters
    ----------
    norm : int
        0 : without normalization
        1 : norm using outer corner of eyes
        2 : norm using related points

    Returns
    -------
    method
        Keras compatible loss function.
    """

    def __init__(self, norm=0, **kwargs):

        self.kwargs = kwargs
        self.norm = norm
        self._norm_dict = {
            0: self._norm_0,
            1: self._norm_1,
            2: self._norm_2, }

    def loss(self, y_true, y_pred):
        """Keras compatible loss function for landmark.

        Notes
        -----
        Input shape should be [batch, num_points * 2]
        e.g., [y0, x0, y1, x1, y2, x2...]
        """
        norm = self._norm_fn(y_true, y_pred)
        num_batches = tf.shape(y_true)[0]
        y_true = tf.reshape(y_true, [num_batches, -1, 2])
        y_pred = tf.reshape(y_pred, [num_batches, -1, 2])
        loss = tf.norm(y_true - y_pred, axis=-1)
        loss = loss / norm
        loss = tf.reduce_mean(loss)
        return loss

    @property
    def _norm_fn(self,):
        return self._norm_dict[self.norm]

    def _norm_0(self, y_true):
        return 1.

    def _norm_1(self, y_true):
        """Norm using outer eye points.

        References
        ----------
        NRMSE: https://link.springer.com/article/10.1186/s13640-018-0324-4
        
        Notes
        -----
        outer eye points should be the first 4 coordinate values.
        e.g., [left_eye_y, left_eye_x, right_eye_y, right_eye_x,...]

        Returns
        -------
        norm : tf.constant
            euclidean distance between outer eye points.
        """
        y0, x0, y1, x1 = self.kwargs['outer_eye_points']
        eye_0 = tf.stack([y_true[:, y0], y_true[:, x0]], -1)
        eye_1 = tf.stack([y_true[:, y1], y_true[:, x1]], -1)
        norm = tf.norm(eye_0 - eye_1, axis=-1)
        norm = tf.expand_dims(norm, -1)
        return norm

    def _norm_2(self, y_true):
        # Norm using pairs of points(EXPERIMENTAL).
        num_batches = tf.shape(y_true)[0]
        y_true = tf.reshape(y_true, [num_batches, -1, 2])
        norm = tf.norm(y_true[:, ::2, :] - y_true[:, 1::2, :], axis=-1)
        norm = tf.concat([tf.tile(norm[:, ::2], [1, 2]),
                          tf.tile(norm[:, 1::2], [1, 2])], -1)
        return norm



"""Face recognition.

Includes face detection, facial landmark detection and facial feature extraction.
Extra Dependencies: tflite-runtime(pip)
"""
from .face_detection import HAARFaceDetector
from .face_detection import LBPFaceDetector
from .face_detection import SSDFaceDetector
from .feature_extraction import FeatureExtractor
from .landmark_detection import LandmarkDetector
from .pipeline import Pipeline




"""Keras implementation of various models."""

import tensorflow as tf
from facelib.dev.layer import Bottleneck
from facelib.dev.image import ImageAugmentor

def mobile_face_net(input_shape=(112,112,3), activation=tf.keras.layers.LeakyReLU):
    """Keras implmentation of MobileFaceNet.
    
    Parameters
    ----------
    input_shape : tuple, optional
        by default (112,112,3)
    activation : [type], optional
        activations used in the model, by default tf.keras.layers.LeakyReLU
    
    Returns
    -------
    [type]
        [description]
    """
    input = tf.keras.layers.Input(input_shape)

    x = tf.keras.layers.Conv2D(
        64, (3,3), (2,2), padding='same')(input)
    x = activation()(x)
    x = tf.keras.layers.BatchNormalization()(x)

    # depthwise conv3x3
    x = tf.keras.layers.DepthwiseConv2D((3, 3), padding='same')(x)
    x = activation()(x)
    x = tf.keras.layers.BatchNormalization()(x)

    # bottleneck
    # t=2, c=64, n=5, s=2
    x = Bottleneck(64, 2, 2, activation)(x)
    x = Bottleneck(64, 2, 1, activation)(x)
    x = Bottleneck(64, 2, 1, activation)(x)
    x = Bottleneck(64, 2, 1, activation)(x)
    x = Bottleneck(64, 2, 1, activation)(x)

    # bottleneck
    # t=4, c=128, n=1, s=2
    x = Bottleneck(128, 4, 2, activation)(x)

    # bottleneck
    # t=2, c=128, n=6, s=1
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)

    # bottleneck
    # t=4, c=128, n=1, s=2
    x = Bottleneck(128, 4, 2, activation)(x)

    # bottleneck
    # t=2, c=128, n=2, s=1
    x = Bottleneck(128, 2, 1, activation)(x)
    x = Bottleneck(128, 2, 1, activation)(x)

    # conv1x1
    x = tf.keras.layers.Conv2D(512, (1, 1), (1, 1), padding='valid')(x)
    x = activation()(x)
    x = tf.keras.layers.BatchNormalization()(x)

    #linear GDConv7x7
    x = tf.keras.layers.DepthwiseConv2D((7, 7), padding='valid')(x)
    x = tf.keras.layers.BatchNormalization()(x)
 
    #linear conv1x1
    x = tf.keras.layers.Conv2D(128, (1, 1), (1, 1), padding='same')(x)
    x = tf.keras.layers.Flatten(name='out_mfn')(x)
    model = tf.keras.models.Model(input, x, name='mobile_face_nets')
    return model


"""A face recognition module with various models.

See:
https://github.com/kutayyildiz/facelib
"""

from setuptools import setup, find_packages 
from os import path
here = path.abspath(path.dirname(__file__))
# 
# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='facelib',
    version='1.1',
    description="""Face Recognition (train/test/deploy)(tensorflow/tflite/keras/edgetpu)""",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/kutayyildiz/facelib',
    author='Kutay YILDIZ',
    author_email='kkutayyildiz@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    package_dir={'facelib': 'facelib'},
    keywords='face,recognition,detection,tensorflow,lite,keras,loss,layer,edgetpu',
    packages=find_packages(),
    python_requires='>=3.5, <3.8',
    install_requires=[
        'joblib',
        'opencv-python',
        'numpy',
        'scikit-image',
        'scikit-learn',
    ],
    extras_require={
        'dev': ['tensorflow'],
    },
    project_urls={
        'Source': 'https://github.com/kutayyildiz/facelib'
    },
    include_package_data=True,
)


"""Script for facial landmark detection."""

from pathlib import Path

import cv2
import numpy as np
import pkg_resources
from tflite_runtime.interpreter import Interpreter, load_delegate

from facelib._utils import helper


class LandmarkDetector:
    def __init__(self, name, input_img_channel='rgb'):
        name_model, inference, tpu = name.split('_')
        tpu = (tpu == 'tpu')
        assert inference in ['fp32', 'int8'], 'Available inferences are: fp32, int8'
        name_model = name_model + '_' + inference
        if tpu:
            delegates = [load_delegate('libedgetpu.so.1')]
            name_model = name_model + '_tpu'
        else:
            name_model = name_model + '_cpu'
            delegates = []
        path_tflite = helper.get_path('landmark_detection', name_model)

        self.landmark_detection_inference = Interpreter(
            model_path = str(path_tflite),
            experimental_delegates = delegates,
        )
        self.landmark_detection_inference.allocate_tensors()
        self.input_index = self.landmark_detection_inference.get_input_details()[0]['index']
        self.output_index = self.landmark_detection_inference.get_output_details()[0]['index']

        assert input_img_channel in ['bgr', 'rgb'], 'Incorrect input_img_channel'
        self.input_img_channel = input_img_channel

    def predict(self, img):
        img = cv2.resize(img, (96, 96))
        if self.input_img_channel == 'bgr':
            img = img[...,::-1]
        img = np.expand_dims(img, 0)
        img = img.astype(np.float32)
        img = (img - 127.5) / 128.
        self.landmark_detection_inference.set_tensor(self.input_index, img)
        self.landmark_detection_inference.invoke()
        landmarks = self.landmark_detection_inference.get_tensor(self.output_index)[0]
        landmarks = landmarks.reshape(-1, 5, 2)
        return landmarks



"""Keras implementation of various layers."""

import tensorflow as tf

class Bottleneck(tf.keras.layers.Layer):
    """Keras implementation of Bottleneck layer.
    
    Parameters
    ----------
    c : int
        number of output channels
    t : int
        expansion rate
    s : int
        strides
    activation : tf.keras.layers.Layer, optional
        activation function compatible with keras
        e.g., tf.keras.layers.LeakyRelu
    
    """

    def __init__(self, c, t, s, activation=tf.keras.layers.ReLU, **kwargs):
        super(Bottleneck, self).__init__(**kwargs)

        # c:output_channels / t:expansion / s:strides
        self.s = s
        tchannel = c * t

        # Input:    h, w, d
        # Operator: 1x1, conv2d
        # Output:   h, w, (c*t)
        self.layer_1 = tf.keras.layers.Conv2D(
            filters = tchannel,
            kernel_size = 1,
            strides=1,
            padding='same',
            use_bias=False)
        self.layer_2 = activation()
        self.layer_3 = tf.keras.layers.BatchNormalization()

        # Input:    h, w, (c*t)
        # Operator: 3x3, dwise, stride=s
        # Output:   h/s, w/s, (c*t)
        self.layer_4 = tf.keras.layers.DepthwiseConv2D(
            kernel_size=(3,3),
            strides=s,
            padding='same',
            depth_multiplier=1,
            use_bias=False)
        self.layer_5 = activation()
        self.layer_6 = tf.keras.layers.BatchNormalization()
        
        # Input:    h/s, w/s, (c*t)
        # Operator: 1x1, conv2d, linear
        # Output:   h/s, w/s, (c)
        self.layer_7 = tf.keras.layers.Conv2D(
            filters=c,
            kernel_size=1,
            padding='same',
            strides=1,
            use_bias=False)
        if s==1:
            self.layer_add = tf.keras.layers.Add()

    def call(self, inputs):
        """Return transformed inputs."""
        x = self.layer_1(inputs)
        x = self.layer_2(x)
        x = self.layer_3(x)
        x = self.layer_4(x)
        x = self.layer_5(x)
        x = self.layer_6(x)
        x = self.layer_7(x)
        if self.s==1:
            x = self.layer_add([x, inputs])
        return x

class DenseL2(tf.keras.layers.Dense):
    """Keras implementation of dense layer with l2 normalization.
    
    Parameters
    ----------
    units : int
        number of output channels
    use_bias : bool
    kernel_regularizer : tf.keras.regularizers.Regularizer
    
    Notes
    -----
    Can be used with margin losses.
    
    """

    def __init__(self, units, use_bias=False, kernel_regularizer=None, **kwargs):
        super(DenseL2, self).__init__(
            units,
            use_bias=use_bias,
            kernel_regularizer=kernel_regularizer,
            **kwargs)

    def call(self, inputs):
        """Return transformed inputs."""
        w_n = tf.nn.l2_normalize(self.kernel, axis=0)
        logits = tf.nn.l2_normalize(inputs, axis=-1)
        out = tf.matmul(logits, w_n)
        return out

class DisturbLabel(tf.keras.layers.Layer):
    """Keras implementation of DisturbLabel."""

    def __init__(self, alpha):
        super(DisturbLabel, self).__init__()
        self.alpha = alpha

    def call(self, y_true):
        """Return transformed inputs.
        
        Notes
        -----
        Should be used on ground truths.
        
        """
        batch_size = tf.shape(y_true)[0]
        num_classes = tf.shape(y_true)[1]
        mask = tf.keras.backend.random_binomial([batch_size], self.alpha)
        mask_r = tf.logical_not(tf.cast(mask, tf.bool))
        mask_r = tf.cast(mask_r, tf.float32)
        mask = tf.stack([mask, mask_r])
        mask = tf.cast(mask, tf.float32)
        y_true_random = tf.random.uniform([batch_size], 0, num_classes, dtype=tf.int32)
        y_true_random = tf.one_hot(y_true_random, num_classes)
        merged = tf.stack([y_true_random, y_true])
        y_true_disturbed = tf.reduce_sum(merged*tf.expand_dims(mask, -1), 0)
        return y_true_disturbed



"""Full face recognition pipeline."""

import cv2
import numpy as np
from skimage import transform as skitrans


class Pipeline:
    """Full face recognition pipeline that can be used
    to detect faces and extract features.

    Parameters
    ----------
    face_detector : Detector
        face detector with a function 'detect'
    landmark_detector : Detector
        landmar detector with a function 'detect'
    face_feature_extractor : Detector
        face feature extractor with a function 'detect'
    
    Examples
    --------
    >>> import cv2
    >>> from facelib import facerec
    >>> face_detector = facerec.SSDFaceDetector()
    >>> landmark_detector = facerec.LandmarkDetector()
    >>> feature_extractor = facerec.FeatureExtractor()
    >>> pipeline = facerec.Pipeline(
            face_detector,
            landmark_detector,
            feature_extractor
        )
    >>> img = cv2.imread('path_to_some_img')[...,::-1]
    >>> bboxes, landmarks, features = fr.predict(img)
    """
        
    def __init__(self, face_detector, face_landmark_detector, face_feature_extractor, scale=None):
        self.face_detector = face_detector
        self.landmark_detector = face_landmark_detector
        self.feature_extractor = face_feature_extractor
        self.scale = scale

    def predict(self, img):
        bboxes = self.face_detector.predict(img)
        if type(bboxes) == tuple:
            bboxes, landmarks = bboxes
            if len(bboxes) == 0:
                return [], [], []
        else:
            if len(bboxes) == 0:
                return [], [], []
            if self.scale is not None:
                bboxes = np.array([self.scale_bbox(bbox, self.scale) for bbox in bboxes])
            faces_cropped = [self.crop_to_bbox(img, bbox) for bbox in bboxes]
            landmarks = np.stack([self.landmark_detector.predict(face) for face in faces_cropped])
            landmarks = np.stack([self.normalize_landmark_coor(img, bbox, landmark) for bbox, landmark in zip(bboxes, landmarks)])
        # do not use np.stack for aligned faces since faces have different width/depth
        faces_aligned = [self.align(img, lm) for lm in landmarks]
        faces_feature = np.stack([self.feature_extractor.predict(fa) for fa in faces_aligned])
        return bboxes, landmarks, faces_feature

    def scale_bbox(self, bbox, scale = [1,1]):
        scale = np.array(scale)
        bbox = bbox.reshape([2, 2])
        height, width = bbox[1] - bbox[0]
        s = (scale - 1) / 2.
        bbox = bbox + [
            [-height * s[0], -width * s[1]], 
            [height * s[0], width * s[1]]]
        return bbox 

    def normalize_landmark_coor(self, img, bbox, landmark):
        # convert landmark coordinates relative to the whole image
        bbox = np.clip(bbox, 0, 1) # some values may be outside of range: [0,1]
        shape_img = img.shape[:2]
        bbox = np.reshape(bbox, [-1, 2]) * shape_img # fraction to pixel
        height, width = bbox[1] - bbox[0]
        landmark = np.reshape(landmark, [-1, 2])
        landmark = landmark * (height, width)
        landmark = landmark + bbox[0]
        landmark = landmark / img.shape[:2]
        return landmark

    def crop_to_bbox(self, img, bbox):
        shape_img = img.shape[:2]
        bbox = np.clip(bbox, 0., 1.)
        bbox = np.reshape(bbox, [-1, 2]) * shape_img
        bbox = bbox.flatten().astype(np.int32)
        img_cropped = img[bbox[0]:bbox[2], bbox[1]:bbox[3], :] # crop
        return img_cropped
    
    def align(self, img, lm):
        lm = np.reshape(lm, [-1, 2])
        lm = lm * img.shape[:2]
        dist = np.linalg.norm(lm[0] - lm[-1]) * 1.5
        dist = dist.astype(np.int32)
        center = np.mean(lm, 0).astype(np.int32)
        bbox = [center - dist, center + dist]
        bbox = np.clip(bbox, 0, img.shape[:2])
        img = img[
            bbox[0][0]: bbox[1][0],
            bbox[0][1]: bbox[1][1],
            :]
        lm = lm - bbox[0]
        lm = lm[...,::-1]
        img_shape = img.shape[:2]
        # coordinates are taken from:
        # https://github.com/deepinsight/insightface/blob/master/src/common/face_preprocess.py
        # normalized to fraction from pixel values
        dst = np.array([
            [0.34191608, 0.46157411],
            [0.65653392, 0.45983393],
            [0.500225  , 0.64050538],
            [0.3709759 , 0.82469198],
            [0.631517  , 0.82325091]])
        dst =  dst * img_shape[::-1]
        tform = skitrans.SimilarityTransform()
        tform.estimate(lm, dst)
        M = tform.params[0:2,:]
        aligned = cv2.warpAffine(img,M,(img_shape[1],img_shape[0]))
        return aligned



"""Keras implementation of various margin losses."""

import tensorflow as tf

class CombinedMargin:
    """Combination of arcface, cosface and sphereface loss functions.

    Notes
    -----
    ref(git) : https://github.com/deepinsight/insightface
    formula : cos(theta * m1 + m2) - m3
    """

    def __init__(self, m1=1., m2=0.3, m3=0.2, scale=64.):
        self.s = tf.constant(scale, tf.float32)
        self.m1 = tf.constant(m1, tf.float32)
        self.m2 = tf.constant(m2, tf.float32)
        self.m3 = tf.constant(m3, tf.float32)

    def loss(self, y_true, y_pred):
        """Keras compatible combined margin loss."""
        original_target_logit = tf.multiply(y_true, y_pred)
        original_target_logit = tf.reduce_sum(original_target_logit, axis=-1)
        original_target_logit = tf.expand_dims(original_target_logit, -1)
        theta = tf.acos(original_target_logit)
        marginal_target_logit = tf.cos(theta * self.m1 + self.m2) - self.m3
        marginal_target_logit = tf.clip_by_value(marginal_target_logit, -0.99, 0.99)
        fc7 = y_pred + y_true * (marginal_target_logit - original_target_logit)
        fc7 = fc7 * self.s
        fc7 = tf.keras.backend.softmax(fc7)
        loss = tf.keras.backend.categorical_crossentropy(y_true, fc7)
        return loss


class ArcFace(CombinedMargin):
    """Arcface loss.

    Notes
    -----
    ref(paper) : https://arxiv.org/abs/1801.07698
    formula : cos(theta + m)
    """

    def __init__(self, m=0.5, scale=64.):
        super().__init__(m1=1., m2=m, m3=0., scale=scale)


class CosineFace(CombinedMargin):
    """CosineFace loss.

    Notes
    -----
    ref(paper) : https://arxiv.org/abs/1801.09414
    formula : cos(theta) - m
    """
    
    def __init__(self, m=0.35, scale=64.):
        super().__init__(m1=1., m2=0., m3=m, scale=scale)


class SphereFace(CombinedMargin):
    """SphereFace loss.

    Notes
    -----
    ref(paper) : https://arxiv.org/abs/1704.08063
    formula : cos(theta * m)
    """
    
    def __init__(self, m=4., scale=1.):
        super().__init__(m1=m, m2=0., m3=0., scale=scale)